package com.gct.dataengine.collector.service.task;

import com.gct.dataengine.collector.entity.BlockData;
import com.gct.dataengine.collector.entity.ConditionMonitorData;
import com.gct.dataengine.collector.entity.vo.MachineServerState;
import com.gct.dataengine.collector.feign.CauldronClient;
import com.gct.dataengine.collector.feign.CentralClient;
import com.gct.dataengine.collector.properties.CollectorProperties;
import com.gct.dataengine.collector.service.task.firstlevel.DeviceStateCollectTask;
import com.gct.dataengine.collector.service.task.firstlevel.TagCollectTask;
import com.gct.dataengine.collector.service.task.firstlevel.ValueCollectTask;
import com.gct.dataengine.collector.service.task.job.ConditionMonitorQuartzJob;
import com.gct.dataengine.collector.service.task.job.DataBlockQuartzJob;
import com.gct.dataengine.collector.service.task.secondlevel.CircleTask;
import com.gct.dataengine.collector.service.task.secondlevel.ConditionMonitorTask;
import com.gct.dataengine.collector.service.task.secondlevel.DataBlockTask;
import com.gct.dataengine.collector.service.task.thirdlevel.TemplateTask;
import com.gct.dataengine.template.entity.enums.HandlerTypeEnum;
import com.gct.dataengine.template.entity.xml.*;
import com.gct.dataengine.template.exception.ConvertFailedException;
import com.gct.dataengine.template.service.ConvertHandler;
import com.gct.dataengine.template.util.DeviceXmlConfigConvert;
import com.gct.driver.lotpi.DataBit;
import com.gct.driver.lotpi.DataBlock;
import com.gct.driver.lotpi.DataType;
import com.gct.driver.lotpi.exception.ReadException;
import com.gct.driver.lotpi.extent.Connector;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.net.Inet4Address;
import java.net.UnknownHostException;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.IllegalFormatConversionException;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * @author xusirui 2022/3/25
 * convert defenition to single template task executor
 */
@Service
@Slf4j
public class TemplateConvert {

    @Autowired
    CollectorProperties collectorProperties;

    @Autowired
    CollectedDataProcessor collectedDataProcessor;

    @Autowired
    CentralClient centralClient;

    /**
     * convert template to Java Object AllDeviceDefinition
     *
     * @param templateId,templatePath
     * @return
     */
    public TemplateTask convert(String templateId,String templatePath) {
        DeviceXmlConfigConvert template = new DeviceXmlConfigConvert(templatePath);
        TemplateDefinition templateDefinition = template.read();
        return convert(templateId,templateDefinition);
    }

    /**
     * convert Java Object AllDeviceDefinition to clear task definition
     *
     * @param templateDefinition
     * @return
     */
    private TemplateTask convert(String templateId,TemplateDefinition templateDefinition) {
        ConnectorPool connectorPool = new ConnectorPool();
        TemplateTask templateTask = new TemplateTask();
        connectorPool.create(templateDefinition.getLotServers());
        //add lotServer to central machine server
        centralClient.addServer(generateMachineServer(templateDefinition.getLotServers(),templateId));
        // add device state collector task
        DeviceStateCollectTask deviceStateCollectTask = monitorCollectTaskProduce(connectorPool,templateId);
        ConditionMonitorTask conditionMonitorTask = new ConditionMonitorTask(collectorProperties.getConditionMonitorInterval(), ConditionMonitorQuartzJob.class, collectedDataProcessor);
        conditionMonitorTask.addTask(deviceStateCollectTask);
        templateTask.addCircleTask(conditionMonitorTask);
        // add data collector task
        for (DataBlockDefinition dataBlockDefinition : templateDefinition.getDataBlockDefinitions()) {
            List<ValueCollectTask> valueCollectTasks = dataValueTaskProduce(dataBlockDefinition.getDataValueDefinitions(), connectorPool);
            List<TagCollectTask> tagCollectTasks = dataTagTaskProduce(dataBlockDefinition.getDataTagDefinitions());
            DataBlockTask dataBlockTask = new DataBlockTask(dataBlockDefinition.getPullInterval(), DataBlockQuartzJob.class,collectedDataProcessor);
            dataBlockTask.addTaskBatch(valueCollectTasks);
            dataBlockTask.addTaskBatch(tagCollectTasks);
            templateTask.addCircleTask(dataBlockTask);
        }
        return templateTask;
    }

    private List<ValueCollectTask> dataValueTaskProduce(DataValueDefinition[] dataValueDefinitions, ConnectorPool connectorPool) {
        return Arrays.stream(dataValueDefinitions).map(dataValueDefinition -> new ValueCollectTask() {

            {
                connector = connectorPool.getById(dataValueDefinition.getSrvId());
                dataBits = Arrays.stream(dataValueDefinition.getDataBitDefinitions()).map(dataBitDefinition ->
                        Objects.nonNull(dataBitDefinition.getAddress()) ?
                                new DataBit(DataType.valueOf(dataBitDefinition.getType()), dataBitDefinition.getAddress()) :
                                new DataBit(DataType.valueOf(dataBitDefinition.getType()), dataBitDefinition.getLowAddress())
                ).collect(Collectors.toList()).toArray(new DataBit[]{});
            }

            public void setDataBlock(DataBlock dataBlock){
                this.dataBlock = dataBlock;
            }

            @Override
            public Object run() {

                Object[] data = null;
                String value = "";
                try {
                    // need combine-pattern
                    data = Arrays.stream(dataBits).map(dataBlock::getData).toArray(); //???
                    //need convert-pattern
                    Class<?> clz = HandlerTypeEnum.getHandlerClassByName(dataValueDefinition.getConvertRef());
                    if (Objects.nonNull(dataValueDefinition.getCombinePattern())) {
                        value = String.format(dataValueDefinition.getCombinePattern(), data);
                    } else if (Objects.nonNull(clz)) {
                        ConvertHandler convertHandler = (ConvertHandler) clz.newInstance();
                        value = convertHandler.convert(data);
                    } else {
                        if (Objects.nonNull(data[0]))
                            value = data[0].toString();
                    }
                } catch (ReadException e) {
                    log.error("read failed, cause by {}", e.getMessage());
                } catch (IllegalFormatConversionException e) {
                    log.error("illegal combine-pattern {},data is {}", dataValueDefinition.getCombinePattern(), data);
                } catch (ConvertFailedException e) {
                    log.error("convert failed {},data is {}", dataValueDefinition.getConvertRef(), data);
                } catch (InstantiationException | IllegalAccessException e) {
                    e.printStackTrace();
                }
                return new BlockData.DataValue(dataValueDefinition.getName(), value, dataValueDefinition.getType());
            }
        }).collect(Collectors.toList());
    }

    private List<TagCollectTask> dataTagTaskProduce(DataTagDefinition[] dataTagDefinitions) {
        return Arrays.stream(dataTagDefinitions).map(dataTagDefinition -> new TagCollectTask() {
            @Override
            public Object run() {
                return new BlockData.DataTag(dataTagDefinition.getName(), dataTagDefinition.getValue());
            }
        }).collect(Collectors.toList());
    }

    private DeviceStateCollectTask monitorCollectTaskProduce(ConnectorPool connectorPool,String templateId){
        return new DeviceStateCollectTask() {
            @Override
            public Object run() {
                ConditionMonitorData state = new ConditionMonitorData();
                state.setTemplateId(templateId);
                //use index map
                for (int i = 0; i < connectorPool.connectors.length; i++) {
                    state.addState(i,connectorPool.getById(i).isOpen());
                }
                return state;
            }
        };
    }

    public List<MachineServerState> generateMachineServer(LotServer[] lotServers, String templateId){
        List<MachineServerState> collect = Arrays.stream(lotServers).map(lotServer -> {
            MachineServerState state = new MachineServerState();
            state.setTemplateId(templateId);
            try {
                state.setCollectorHost(Inet4Address.getLocalHost().getHostAddress());
            } catch (UnknownHostException e) {
                e.printStackTrace();
            }
            state.setCollectorId(collectorProperties.getId());
            state.setCollectorName(collectorProperties.getName());
            state.setCreateTime(LocalDateTime.now());
            state.setDeviceName(lotServer.getName());
            state.setOnline(0);
            state.setProtocol(lotServer.getUrl().split("://")[0]);
            state.setIpAddress(lotServer.getUrl().split("://")[1].split("\\?")[0].split(":")[0]);
            state.setPort(lotServer.getUrl().split("://")[1].split("\\?")[0].split(":")[1]);
            state.setUpdateTime(LocalDateTime.now());
            return state;
        }).collect(Collectors.toList());
        for (int i = 0; i < collect.size(); i++) {
            collect.get(i).setDeviceId(Integer.toString(i));
        }
        return collect;

    }
}
